import asyncio
import logging

import i2plib.sam
import i2plib.exceptions
import i2plib.utils

BUFFER_SIZE = 65536

def parse_reply(data):
    return i2plib.sam.Answer(data.decode().strip())

async def get_sam_socket(sam_address=i2plib.sam.DEFAULT_ADDRESS, loop=None):
    """A couroutine used to create a new SAM socket.

    :param sam_address: (optional) SAM API address
    :param loop: (optional) event loop instance
    :return: A (reader, writer) pair
    """
    reader, writer = await asyncio.open_connection(*sam_address, loop=loop)
    writer.write(i2plib.sam.hello("3.1", "3.1"))
    reply = parse_reply(await reader.read(BUFFER_SIZE))
    if reply.ok:
        return (reader, writer)
    else:
        raise i2plib.exceptions.SAM_EXCEPTIONS[reply["RESULT"]]()

async def dest_lookup(domain, sam_address=i2plib.sam.DEFAULT_ADDRESS, 
                      loop=None):
    """A coroutine used to lookup a full I2P destination by .i2p domain or 
    .b32.i2p address.

    :param domain: Address to be resolved, can be a .i2p domain or a .b32.i2p 
                   address.
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :return: An instance of :class:`i2plib.Destination`
    """
    reader, writer = await get_sam_socket(sam_address, loop)
    writer.write(i2plib.sam.naming_lookup(domain))
    reply = parse_reply(await reader.read(BUFFER_SIZE))
    if reply.ok:
        writer.close()
        return i2plib.sam.Destination(reply["VALUE"])
    else:
        raise i2plib.exceptions.SAM_EXCEPTIONS[reply["RESULT"]]()

async def new_destination(sam_address=i2plib.sam.DEFAULT_ADDRESS, loop=None,
                      sig_type=i2plib.sam.Destination.default_sig_type):
    """A coroutine used to generate a new destination with a private key of a 
    chosen signature type.

    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :param sig_type: (optional) Signature type
    :return: An instance of :class:`i2plib.Destination`
    """
    reader, writer = await get_sam_socket(sam_address, loop)
    writer.write(i2plib.sam.dest_generate(sig_type))
    reply = parse_reply(await reader.read(BUFFER_SIZE))
    writer.close()
    return i2plib.sam.Destination(reply["PRIV"], has_private_key=True)

async def create_session(session_name, sam_address=i2plib.sam.DEFAULT_ADDRESS, 
                         loop=None, style="STREAM",
                         signature_type=i2plib.sam.Destination.default_sig_type,
                         destination=None, options={}):
    """A coroutine used to create a new SAM session.

    :param session_name: Session nick name
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :param style: (optional) Session style, can be STREAM, DATAGRAM, RAW
    :param signature_type: (optional) If the destination is TRANSIENT, this 
                        signature type is used
    :param destination: (optional) Destination to use in this session. Can be 
                        a base64 encoded string, :class:`i2plib.Destination` 
                        instance or None. TRANSIENT destination is used when it
                        is None.
    :param options: (optional) A dict object with i2cp options
    :return: A (reader, writer) pair
    """
    logging.debug("Creating session {}".format(session_name))
    if destination:
        if type(destination) == i2plib.sam.Destination:
            destination = destination
        else:
            destination = i2plib.sam.Destination(
                    destination, has_private_key=True)

        dest_string = destination.private_key.base64
    else:
        dest_string = i2plib.sam.TRANSIENT_DESTINATION

    options = " ".join(["{}={}".format(k, v) for k, v in options.items()])

    reader, writer = await get_sam_socket(sam_address, loop)
    writer.write(i2plib.sam.session_create(
            style, session_name, dest_string, options))

    reply = parse_reply(await reader.read(BUFFER_SIZE))
    if reply.ok:
        if not destination:
            destination = i2plib.sam.Destination(
                    reply["DESTINATION"], has_private_key=True) 
        logging.debug(destination.base32)
        logging.debug("Session created {}".format(session_name))
        return (reader, writer)
    else:
        raise i2plib.exceptions.SAM_EXCEPTIONS[reply["RESULT"]]()

async def stream_connect(session_name, destination, 
                         sam_address=i2plib.sam.DEFAULT_ADDRESS, loop=None):
    """A coroutine used to connect to a remote I2P destination.

    :param session_name: Session nick name
    :param destination: I2P destination to connect to
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :return: A (reader, writer) pair
    """
    logging.debug("Connecting stream {}".format(session_name))
    if isinstance(destination, str) and not destination.endswith(".i2p"):
        destination = i2plib.sam.Destination(destination)
    elif isinstance(destination, str):
        destination = await dest_lookup(destination, sam_address, loop)

    reader, writer = await get_sam_socket(sam_address, loop)
    writer.write(i2plib.sam.stream_connect(session_name, destination.base64,
                                           silent="false"))
    reply = parse_reply(await reader.read(BUFFER_SIZE))
    if reply.ok:
        logging.debug("Stream connected {}".format(session_name))
        return (reader, writer)
    else:
        raise i2plib.exceptions.SAM_EXCEPTIONS[reply["RESULT"]]()

async def stream_accept(session_name, sam_address=i2plib.sam.DEFAULT_ADDRESS,
                        loop=None):
    """A coroutine used to accept a connection from the I2P network.

    :param session_name: Session nick name
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :return: A (reader, writer) pair
    """
    reader, writer = await get_sam_socket(sam_address, loop)
    writer.write(i2plib.sam.stream_accept(session_name, silent="false"))
    reply = parse_reply(await reader.read(BUFFER_SIZE))
    if reply.ok:
        return (reader, writer)
    else:
        raise i2plib.exceptions.SAM_EXCEPTIONS[reply["RESULT"]]()

### Context managers

class Session:
    """Async SAM session context manager.

    :param session_name: Session nick name
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :param style: (optional) Session style, can be STREAM, DATAGRAM, RAW
    :param signature_type: (optional) If the destination is TRANSIENT, this 
                        signature type is used
    :param destination: (optional) Destination to use in this session. Can be 
                        a base64 encoded string, :class:`i2plib.Destination` 
                        instance or None. TRANSIENT destination is used when it
                        is None.
    :param options: (optional) A dict object with i2cp options
    :return: :class:`i2plib.Session` object
    """
    def __init__(self, session_name, sam_address=i2plib.sam.DEFAULT_ADDRESS, 
                         loop=None, style="STREAM",
                         signature_type=i2plib.sam.Destination.default_sig_type,
                         destination=None, options={}):
        self.session_name = session_name
        self.sam_address = sam_address
        self.loop = loop
        self.style = style
        self.signature_type = signature_type
        self.destination = destination
        self.options = options

    async def __aenter__(self):
        self.reader, self.writer = await create_session(self.session_name, 
                sam_address=self.sam_address, loop=self.loop, style=self.style, 
                signature_type=self.signature_type, 
                destination=self.destination, options=self.options)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        ### TODO handle exceptions
        self.writer.close()

class StreamConnection:
    """Async stream connection context manager.

    :param session_name: Session nick name
    :param destination: I2P destination to connect to
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :return: :class:`i2plib.StreamConnection` object
    """
    def __init__(self, session_name, destination, 
                 sam_address=i2plib.sam.DEFAULT_ADDRESS, loop=None):
        self.session_name = session_name
        self.sam_address = sam_address
        self.loop = loop
        self.destination = destination

    async def __aenter__(self):
        self.reader, self.writer = await stream_connect(self.session_name, 
                self.destination, sam_address=self.sam_address, loop=self.loop)
        self.read = self.reader.read
        self.write = self.writer.write
        return self

    async def __aexit__(self, exc_type, exc, tb):
        ### TODO handle exceptions
        self.writer.close()

class StreamAcceptor:
    """Async stream acceptor context manager.

    :param session_name: Session nick name
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :return: :class:`i2plib.StreamAcceptor` object
    """
    def __init__(self, session_name, sam_address=i2plib.sam.DEFAULT_ADDRESS, 
                 loop=None):
        self.session_name = session_name
        self.sam_address = sam_address
        self.loop = loop

    async def __aenter__(self):
        self.reader, self.writer = await stream_accept(self.session_name, 
                        sam_address=self.sam_address, loop=self.loop)
        self.read = self.reader.read
        self.write = self.writer.write
        return self

    async def __aexit__(self, exc_type, exc, tb):
        ### TODO handle exceptions
        self.writer.close()



import socket
from base64 import b64decode, b64encode, b32encode
from hashlib import sha256
import struct
import random
import string
import re

from .exceptions import SAM_EXCEPTIONS

I2P_B64_CHARS = "-~"

def i2p_b64encode(x):
    """Encode I2P destination"""
    return b64encode(x, altchars=I2P_B64_CHARS.encode()).decode() 

def i2p_b64decode(x):
    """Decode I2P destination"""
    return b64decode(x, altchars=I2P_B64_CHARS, validate=True)

SAM_BUFSIZE = 4096
DEFAULT_ADDRESS = ("127.0.0.1", 7656)
DEFAULT_MIN_VER = "3.1"
DEFAULT_MAX_VER = "3.1"
TRANSIENT_DESTINATION = "TRANSIENT"

VALID_BASE32_ADDRESS = re.compile(r"^([a-zA-Z0-9]{52}).b32.i2p$")
VALID_BASE64_ADDRESS = re.compile(r"^([a-zA-Z0-9-~=]{516,528})$")

class Answer(object):
    """Parse answer from SAM bridge to an object"""
    def __init__(self, s):
        self.opts = {}
        if type(s) != str:
            self._reply_string = s.decode().strip()
        else:
            self._reply_string = s

        self.cmd, opts = self._reply_string.split(" ", 1)
        for v in opts.split(" ")[1:]:
            data = v.split("=", 1) if "=" in v else (v, True)
            self.opts[data[0]] = data[1]

    def __getitem__(self, key):
        return self.opts[key]

    @property
    def ok(self):
        return self["RESULT"] == "OK"

    def __repr__(self):
        return self._reply_string

def get_socket(sam_address=None):
    """Return new SAM socket"""
    sam_address = sam_address or DEFAULT_ADDRESS
    sam_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sam_socket.connect(sam_address)
    sam_socket.send(hello(DEFAULT_MIN_VER, DEFAULT_MAX_VER))
    a = get_response(sam_socket)
    if a.ok:
        return sam_socket
    else:
        raise SAM_EXCEPTIONS[a["RESULT"]]()

def lookup(sam_socket, name):
    """Lookup destination by name"""
    sam_socket.send(naming_lookup(name))
    a = get_response(sam_socket)
    if a.ok:
        return Destination(a["VALUE"])
    else:
        raise SAM_EXCEPTIONS[a["RESULT"]]()

def generate_session_id(length=6):
    """Generate random session id"""
    rand = random.SystemRandom()
    sid = [rand.choice(string.ascii_letters) for _ in range(length)]
    return "i2plib-" + "".join(sid)

def get_response(sam_socket):
    """Read answer from SAM API"""
    return Answer(sam_socket.recv(SAM_BUFSIZE))


# SAM request messages

def hello(min_version, max_version):
    return "HELLO VERSION MIN={} MAX={}\n".format(min_version, 
            max_version).encode()

def session_create(style, session_id, destination, options=""):
    return "SESSION CREATE STYLE={} ID={} DESTINATION={} {}\n".format(
            style, session_id, destination, options).encode()


def stream_connect(session_id, destination, silent="false"):
    return "STREAM CONNECT ID={} DESTINATION={} SILENT={}\n".format(
            session_id, destination, silent).encode()

def stream_accept(session_id, silent="false"):
    return "STREAM ACCEPT ID={} SILENT={}\n".format(session_id, silent).encode()

def stream_forward(session_id, port, options=""):
    return "STREAM FORWARD ID={} PORT={} {}\n".format(
            session_id, port, options).encode()



def naming_lookup(name):
    return "NAMING LOOKUP NAME={}\n".format(name).encode()

def dest_generate(signature_type):
    return "DEST GENERATE SIGNATURE_TYPE={}\n".format(signature_type).encode()

class Destination(object):
    """I2P destination

    https://geti2p.net/spec/common-structures#destination

    :param data: (optional) Base64 encoded data or binary data 
    :param path: (optional) A path to a file with binary data 
    :param has_private_key: (optional) Does data have a private key? 
    """

    ECDSA_SHA256_P256 = 1
    ECDSA_SHA384_P384 = 2
    ECDSA_SHA512_P521 = 3
    EdDSA_SHA512_Ed25519 = 7

    default_sig_type = EdDSA_SHA512_Ed25519

    _pubkey_size = 256
    _signkey_size = 128
    _min_cert_size = 3

    def __init__(self, data=None, path=None, has_private_key=False):
        #: Binary destination
        self.data = bytes() 
        #: Base64 encoded destination
        self.base64 = ""    
        #: :class:`i2plib.PrivateKey` instance or None
        self.private_key = None    
        
        if path:
            with open(path, "rb") as f: data = f.read()

        if data and has_private_key:
            self.private_key = PrivateKey(data)

            cert_len = struct.unpack("!H", self.private_key.data[385:387])[0]
            data = self.private_key.data[:387+cert_len]

        if not data:
            raise Exception("Can't create a destination with no data")

        self.data = data if type(data) == bytes else i2p_b64decode(data)
        self.base64 = data if type(data) == str else i2p_b64encode(data)

    def __repr__(self):
        return "<Destination: {}>".format(self.base32)

    @property
    def base32(self):
        """Base32 destination hash of this destination"""
        desthash = sha256(self.data).digest()
        return b32encode(desthash).decode()[:52].lower()
    
class PrivateKey(object):
    """I2P private key

    https://geti2p.net/spec/common-structures#keysandcert

    :param data: Base64 encoded data or binary data 
    """

    def __init__(self, data):
        #: Binary private key
        self.data = data if type(data) == bytes else i2p_b64decode(data)
        #: Base64 encoded private key
        self.base64 = data if type(data) == str else i2p_b64encode(data)

class StreamSession(object):

    def __init__(self, sam_address=DEFAULT_ADDRESS, \
            session_id=None, destination=TRANSIENT_DESTINATION, options="", \
            min_ver=DEFAULT_MIN_VER, max_ver=DEFAULT_MAX_VER):
        self.sam_address = sam_address
        self.min_ver = min_ver
        self.max_ver = max_ver
        self.session_id = session_id or generate_session_id()
        self._session_socket = get_socket(self.sam_address)
        self._session_socket.send(
            session_create("STREAM", self.session_id, destination, options))
        a = get_response(self._session_socket)

        if a.ok:
            self.destination = a["DESTINATION"]
        else:
            raise SAM_EXCEPTIONS[a["RESULT"]]()

    def _get_socket(self):
        """New socket for this session"""
        return get_socket(self.sam_address)

    def connect(self, dest):
        """Return new I2P stream to destination"""
        sam_socket = self._get_socket()
        sam_socket.send(stream_connect(self.session_id, dest, silent="false"))
        a = get_response(sam_socket)
        if a.ok:
            return sam_socket
        else:
            raise SAM_EXCEPTIONS[a["RESULT"]]()

    def accept(self):
        """Return new accepting socket"""
        sam_socket = self._get_socket()
        sam_socket.send(stream_accept(self.session_id, silent="false"))
        return sam_socket

    def forward(self, session_id, port, options=""):
        sam_socket = self._get_socket()
        sam_socket.send(stream_forward(self.session_id, port, options))
        return sam_socket




import socket
import os
import i2plib.sam

def get_free_port():
    """Get a free port on your local host"""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind(('', 0))
    free_port = s.getsockname()[1]
    s.close()
    return free_port

def is_address_accessible(address):
    """Check if address is accessible or down"""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    is_accessible = s.connect_ex(address) == 0
    s.close()
    return is_accessible

def address_from_string(address_string):
    """Address tuple from host:port string"""
    address = address_string.split(":")
    return (address[0], int(address[1]))

def get_sam_address():
    """
    Get SAM address from environment variable I2P_SAM_ADDRESS, or use a default
    value
    """
    value = os.getenv("I2P_SAM_ADDRESS")
    return address_from_string(value) if value else i2plib.sam.DEFAULT_ADDRESS

def get_new_destination(sam_address=i2plib.sam.DEFAULT_ADDRESS, 
                        sig_type=i2plib.sam.Destination.default_sig_type):
    """Generates new I2P destination of a chosen signature type"""
    sam_socket = i2plib.sam.get_socket(sam_address)
    sam_socket.send(i2plib.sam.dest_generate(sig_type))
    a = i2plib.sam.get_response(sam_socket)
    sam_socket.close()
    return i2plib.sam.Destination(a['PRIV'], has_private_key=True)




import unittest
import os

import i2plib.sam

TEST_KEY = os.path.join(os.getcwd(), 'test', 'data', 'test.dat')
TEST_DEST_B32 = "testvjtq2sfj5a6m7dwecpuxwbpwglgrma4i2iisxg2g5bnogm5a"
TEST_DEST_B64 = "Q8VZ41jSRFutjRYDbnDioTfMitFMAMINRqrgcwISQ3ObOyzjBX9Kz~OJz7ShyGqdILTxgeDYkGTpYEp4HIleUHwXUVaX8tEjEYQPYMTZr-ol-biPwokefgUtOG5MKB7d3pPN9z5j6wFaYfUhSuXq6eVkMqxE9S3MGP5D8l2ihYhvlKrL5uTW0up8GnO~o0e7SXoYRUES1TuA78GkB10qOJKEW4ow0YFr2m1I7gL7mrJu44XfvAALOmyKuJxvXxpevAeeW1TWEWONP9skUC9N3mvQsGRBCzIeZEgSlT0Vkdg8e5r8nUYAWxZmeY4U-vW2lchCb4hz4Tnuj9nWsMdj4bjnuBe51SE4G93Fw73o3w4PHL2blh0yVLGOlVzHOv6kz2TPg4dnjioI0sFTnFv4CyARPHHJtKMQi7JF6SucMSO4F2u~XdSIMnV845OIL1W52097~LqX9uW7sg-pbueB1FSUIyv4GpwZo2XKxSm6EQ5tv2tYs9LowPDQzPZhXAAUBQAEAAcAAA=="

class TestDestination(unittest.TestCase):

    def test_private_key(self):
        pk = i2plib.sam.Destination(path=TEST_KEY, has_private_key=True)
        self.assertEqual(pk.base32, TEST_DEST_B32)
        self.assertEqual(pk.base64, TEST_DEST_B64)

    def test_destination(self):
        dest = i2plib.sam.Destination(TEST_DEST_B64)
        self.assertEqual(dest.base32, TEST_DEST_B32)
